#include <avr/io.h>
#include <string.h>

#include "mn2fw.h"
#include "mn2lcd.h"

/*
 * LCD interface.
 */
#define LCD_DATA_PORT   PORTD
#define LCD_DATA_DDR    DDRD
#define LCD_DATA_BIT    0

#define LCD_E_PORT      PORTC
#define LCD_E_DDR       DDRC
#define LCD_E_MASK      0x02

#define LCD_RW_PORT     PORTC
#define LCD_RW_DDR      DDRC
#define LCD_RW_MASK     0x04

#define LCD_RS_PORT     PORTC
#define LCD_RS_DDR      DDRC
#define LCD_RS_MASK     0x08

#if 0
#define LCD_CS1_PORT    PORTB
#define LCD_CS1_DDR     DDRB
#define LCD_CS1_BIT     7

#define LCD_CS2_PORT    PORTB
#define LCD_CS2_DDR     DDRB
#define LCD_CS2_BIT     6
#endif

static unsigned char lcd_rows;
static unsigned char lcd_cols;

void LcdDelay(unsigned long count)
{
    while (count--) {
        asm volatile ("nop");
    }
}

void LcdWriteCmd(unsigned char ch)
{
    LCD_DATA_DDR = 0xFF;
    LCD_RW_PORT &= ~LCD_RW_MASK;
    LCD_RS_PORT &= ~LCD_RS_MASK;
    LCD_DATA_PORT = ch;

    LCD_E_PORT |= LCD_E_MASK;
    LcdDelay(32);
    LCD_E_PORT &= ~LCD_E_MASK;
    LcdDelay(32);
}

void LcdWriteData(unsigned char ch)
{
    LCD_DATA_DDR = 0xFF;
    LCD_RW_PORT &= ~LCD_RW_MASK;
    LCD_RS_PORT |= LCD_RS_MASK;
    LCD_DATA_PORT = ch;

    LCD_E_PORT |= LCD_E_MASK;
    LcdDelay(32);
    LCD_E_PORT &= ~LCD_E_MASK;
    LcdDelay(32);
}

unsigned char LcdReadStatus(void)
{
    unsigned char rc;

    LCD_DATA_DDR = 0x00;
    LCD_DATA_PORT = 0xFF;
    LCD_RW_PORT |= LCD_RW_MASK;
    LCD_RS_PORT &= ~LCD_RS_MASK;

    LCD_E_PORT |= LCD_E_MASK;
    LcdDelay(32);
    rc = LCD_DATA_PORT;
    LCD_E_PORT &= ~LCD_E_MASK;
    LcdDelay(32);

    return rc;
}

void LcdWaitReady(void)
{
    while ((LcdReadStatus() & 0x80) == 0);
}

int LcdSetPos(unsigned char row, unsigned char col)
{
    unsigned char pos;

    if (lcd_rows == 0 || lcd_cols == 0) {
        return -1;
    }

    if (row >= lcd_rows) {
        row = lcd_rows;
    }
    else if (col >= lcd_cols) {
        col = lcd_cols - 1;
    }

    if (row == 0) {
        pos = col;
    }
    else if (row == 1) {
        pos = 64 + col;
    }
    else {
        pos = (row & 1) * 64 + lcd_cols + col;
    }
    LcdWriteCmd(LCD_CMD_DDRAM | pos);

    return 0;
}

/*
 * Set LCD user definable characters.
 */
static void LcdSetSpecials(void)
{
    unsigned char spec_char[4][8] = {
        {0x1F, 0x00, 0x0E, 0x01, 0x0F, 0x11, 0x0F, 0x00},   /*  */
        {0x0E, 0x00, 0x0E, 0x11, 0x1F, 0x10, 0x0E, 0x00},   /*  */
        {0x04, 0x00, 0x0E, 0x11, 0x11, 0x11, 0x0E, 0x00},   /*  */
        {0x04, 0x00, 0x11, 0x11, 0x11, 0x13, 0x0D, 0x00}    /*  */
    };
    unsigned char i;
    unsigned char j;
    unsigned char code;

    for (i = 0; i < 4; i++) {
        for (j = 0; j < 8; j++) {
            code = 0x40 | ((i + 1) << 3) | j;
            LcdWriteCmd(code);
            LcdWaitReady();
            LcdWriteData(spec_char[i][j]);
            LcdWaitReady();
        }
    }
}

int LcdInit(unsigned char mode)
{
    unsigned long t;

    for (t = 0; t < 0x770; t++) {
        asm volatile ("nop");
    }                           //15 ms

    if (mode == LCD_MODE_1X8) {
        lcd_rows = 1;
        lcd_cols = 8;
    }
    else if (mode == LCD_MODE_1X16) {
        lcd_rows = 1;
        lcd_cols = 16;
    }
    else if (mode == LCD_MODE_1X20) {
        lcd_rows = 1;
        lcd_cols = 20;
    }
    else if (mode == LCD_MODE_2X8) {
        lcd_rows = 2;
        lcd_cols = 8;
    }
    else if (mode == LCD_MODE_2X16) {
        lcd_rows = 2;
        lcd_cols = 16;
    }
    else if (mode == LCD_MODE_2X20) {
        lcd_rows = 2;
        lcd_cols = 20;
    }
    else if (mode == LCD_MODE_2X40) {
        lcd_rows = 2;
        lcd_cols = 40;
    }
    else if (mode == LCD_MODE_4X16) {
        lcd_rows = 4;
        lcd_cols = 16;
    }
    else if (mode == LCD_MODE_4X20) {
        lcd_rows = 4;
        lcd_cols = 20;
    }
    else {
        lcd_rows = 0;
        lcd_cols = 0;

        return -1;
    }

    LCD_DATA_PORT = 0x00;
    LCD_DATA_DDR = 0xFF;

    LCD_RS_PORT &= ~LCD_RS_MASK;
    LCD_RS_DDR |= LCD_RS_MASK;

    LCD_RW_PORT &= ~LCD_RW_MASK;
    LCD_RW_DDR |= LCD_RW_MASK;

    LCD_E_PORT |= LCD_E_MASK;
    LCD_E_DDR |= LCD_E_MASK;

    /* LCD function set. */
    if (lcd_rows == 1) {
        LcdWriteCmd(LCD_CMD_FUNCTION | LCD_CMD_FUNCTION_8BIT);
    }
    else {
        LcdWriteCmd(LCD_CMD_FUNCTION | LCD_CMD_FUNCTION_8BIT | LCD_CMD_FUNCTION_2LINES);
    }
    LcdWaitReady();

    /* Switch LCD on. */
    LcdWriteCmd(LCD_CMD_ON_CTRL | LCD_CMD_ON_DISPLAY);
    LcdWaitReady();

    /* Set entry mode. */
    LcdWriteCmd(LCD_CMD_ENTRY_MODE | LCD_CMD_ENTRY_INC);
    LcdWaitReady();

    /* Set move mode. */
    LcdWriteCmd(LCD_CMD_MOVE);
    LcdWaitReady();

    /* Clear display. */
    LcdWriteCmd(LCD_CMD_CLR);
    LcdWaitReady();

    /* Send cursor home. */
    LcdWriteCmd(LCD_CMD_HOME);
    LcdWaitReady();

    LcdSetSpecials();

    return 0;
}

