/*!
 * Copyright (C) 2007 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * $Log$
 */

/*!
 * \example progm168/progm168.c
 */
#include <cfg/os.h>
#include <dev/board.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <errno.h>

#include <dev/urom.h>
#include <dev/nplmmc.h>

#include <sys/thread.h>
#include <sys/timer.h>
#include <sys/event.h>
#include <sys/bankmem.h>
#include <sys/heap.h>
#include <dev/irqreg.h>

#include <fs/phatfs.h>

#include <mn2ctrl.h>

#include <dev/vs10xx.h>

#include <dev/avrtarget.h>
#include "media2spi.h"

#define MEDIA2_FIRMWARE_FILE    "UROM:mn2fw.img"

#define MUSIC_ROOT_DIR          "PHAT0:/"

#ifndef MEDIA2_FUSES
#define MEDIA2_FUSES    0x00F9DFE2UL
#endif

u_char data[AVRTARGET_PAGESIZE];

static int Media2Firmware(int fd, ureg_t mode)
{
    int rc = -1;
    u_long page = 0;
    u_long signature;
    ureg_t i;
    int got;

    /* Initialize serial peripheral interface. */
    if (AvrTargetInit()) {
        if (mode & 2) {
            puts("Error: Bad SPI configuration.");
        }
        return rc;
    }

    /* Activate the interface. */
    AvrTargetSelect(1);

    /*
     * Try to enter programming mode. 
     */
    for (i = 0; i < 10; i++) {
        AvrTargetReset(1);
        NutDelay(1);
        if (AvrTargetProgEnable() == 0) {
            break;
        }
        AvrTargetReset(0);
        NutDelay(1);
    }
    if (i >= 10) {
        if (mode & 2) {
            puts("Error: Failed to enter programming mode.");
        }
    }

    /*
     * Make sure we got the right device and erase it.
     */
    else if ((signature = AvrTargetSignature()) != AVRSIGNATURE_MEGA168) {
        if (mode & 2) {
            printf("Error: Bad signature %06lX\n", signature);
        }
    } 
    else if ((mode & 1) != 0 && AvrTargetChipErase()) {
        if (mode & 2) {
            puts("Error: Erasing chip failed.");
        }
    }

    /*
     * Upload the image.
     */
    else {
        for (page = 0;; page++) {
            if (mode & 2) {
                printf("%s page %lu\n", (mode & 1) ? "Flashing" : "Verifying", page);
            }
            if ((got = _read(fd, data, AVRTARGET_PAGESIZE)) != 0) {
                if (got < 0) {
                    if (mode & 2) {
                        printf("Error %d: Reading image file.\n", errno);
                    }
                    break;
                }
                if (got < AVRTARGET_PAGESIZE) {
                    memset(data + got, 0xFF, AVRTARGET_PAGESIZE - got);
                }
                if (mode & 1) {
                    AvrTargetPageLoad(page, data);
                    if (AvrTargetPageWrite(page)) {
                        if (mode & 2) {
                            puts("Error: Memory programming failed.");
                        }
                        break;
                    }
                } else if (AvrTargetPageVerify(page, data)) {
                    if (mode & 2) {
                        puts("Error: Memory verify failed.");
                    }
                    break;
                }
            }
            if (got < AVRTARGET_PAGESIZE) {
                rc = 0;
                break;
            }
        }
    }
    if (rc == 0) {
        /* Check and optionally update the fuses. */
        if ((mode & 1) != 0 && AvrTargetFusesRead() != MEDIA2_FUSES) {
            if (mode & 2) {
                puts("Updating fuses.");
            }
            AvrTargetFusesWriteSafe(MEDIA2_FUSES);
        }
        AvrTargetReset(0);
    }
    /* De-activate the interface. */
    AvrTargetSelect(0);

    return rc;
}

/*!
 * \brief Check, upload and verify Medianut 2 firmware.
 */
int Medianut2Upload(char *path)
{
    int fd;

    /*
     * Open the firmware image file.
     */
    if ((fd = _open(path, _O_RDONLY | _O_BINARY)) == -1) {
        printf("Error %d: Failed to open %s.\n", errno, path);
        return -1;
    }

    printf("Checking firmware...");
    if (Media2Firmware(fd, 0)) {
        printf("not found\nUpdating...");
        _seek(fd, 0, SEEK_SET);
        if (Media2Firmware(fd, 1)) {
            puts("failed");
        } else {
            printf("OK\nVerifying...");
            _seek(fd, 0, SEEK_SET);
            if (Media2Firmware(fd, 0)) {
                puts("failed");
            } else {
                puts("OK");
            }
        }
    } else {
        puts("OK");
    }
    _close(fd);

    return 0;
}
