/*!
 * Copyright (C) 2007 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * $Log$
 */

/*!
 * \example progm168/progm168.c
 */
#include <cfg/os.h>
#include <dev/board.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <io.h>
#include <fcntl.h>
#include <errno.h>

#include <dev/urom.h>
#include <dev/nplmmc.h>

#include <sys/thread.h>
#include <sys/timer.h>
#include <sys/event.h>
#include <sys/bankmem.h>
#include <sys/heap.h>
#include <dev/irqreg.h>

#include <fs/phatfs.h>

#include <mn2ctrl.h>

#include <dev/vs10xx.h>

#include <dev/avrtarget.h>
#include "media2spi.h"

#define MEDIA2_FIRMWARE_FILE    "UROM:mn2fw.img"

#define MUSIC_ROOT_DIR          "PHAT0:/"

#ifndef MEDIA2_FUSES
#define MEDIA2_FUSES    0x00F9DFE2UL
#endif

u_char data[AVRTARGET_PAGESIZE];

int Media2Firmware(int fd, ureg_t mode)
{
    int rc = -1;
    u_long page = 0;
    u_long signature;
    ureg_t i;
    int got;

    /* Initialize serial peripheral interface. */
    if (AvrTargetInit()) {
        if (mode & 2) {
            puts("Error: Bad SPI configuration.");
        }
        return rc;
    }

    /* Activate the interface. */
    AvrTargetSelect(1);

    /*
     * Try to enter programming mode. 
     */
    for (i = 0; i < 10; i++) {
        AvrTargetReset(1);
        NutDelay(1);
        if (AvrTargetProgEnable() == 0) {
            break;
        }
        AvrTargetReset(0);
        NutDelay(1);
    }
    if (i >= 10) {
        if (mode & 2) {
            puts("Error: Failed to enter programming mode.");
        }
    }

    /*
     * Make sure we got the right device and erase it.
     */
    else if ((signature = AvrTargetSignature()) != AVRSIGNATURE_MEGA168) {
        if (mode & 2) {
            printf("Error: Bad signature %06lX\n", signature);
        }
    } 
    else if ((mode & 1) != 0 && AvrTargetChipErase()) {
        if (mode & 2) {
            puts("Error: Erasing chip failed.");
        }
    }

    /*
     * Upload the image.
     */
    else {
        for (page = 0;; page++) {
            if (mode & 2) {
                printf("%s page %lu\n", (mode & 1) ? "Flashing" : "Verifying", page);
            }
            if ((got = _read(fd, data, AVRTARGET_PAGESIZE)) != 0) {
                if (got < 0) {
                    if (mode & 2) {
                        printf("Error %d: Reading image file.\n", errno);
                    }
                    break;
                }
                if (got < AVRTARGET_PAGESIZE) {
                    memset(data + got, 0xFF, AVRTARGET_PAGESIZE - got);
                }
                if (mode & 1) {
                    AvrTargetPageLoad(page, data);
                    if (AvrTargetPageWrite(page)) {
                        if (mode & 2) {
                            puts("Error: Memory programming failed.");
                        }
                        break;
                    }
                } else if (AvrTargetPageVerify(page, data)) {
                    if (mode & 2) {
                        puts("Error: Memory verify failed.");
                    }
                    break;
                }
            }
            if (got < AVRTARGET_PAGESIZE) {
                rc = 0;
                break;
            }
        }
    }
    if (rc == 0) {
        /* Check and optionally update the fuses. */
        if ((mode & 1) != 0 && AvrTargetFusesRead() != MEDIA2_FUSES) {
            if (mode & 2) {
                puts("Updating fuses.");
            }
            AvrTargetFusesWriteSafe(MEDIA2_FUSES);
        }
        AvrTargetReset(0);
    }
    /* De-activate the interface. */
    AvrTargetSelect(0);

    return rc;
}

/*
 * Play MP3 file from local file system.
 *
 * \param path Pathname of the MP3 file to play.
 *
 * \return 0 on success, -1 if opening the file failed.
 */
static int PlayMp3File(char *path)
{
    int fd;
    size_t rbytes;
    u_char *mp3buf;
    int got;
    u_char ief;

    /*
     * Open the MP3 file.
     */
    printf("Play %s: ", path);
    if ((fd = _open(path, _O_RDONLY | _O_BINARY)) == -1) {
        printf("Error %d\n", errno);
        return -1;
    }
    puts("OK");

    /* 
     * Reset decoder buffer.
     */
    ief = VsPlayerInterrupts(0);
    NutSegBufReset();
    VsPlayerInterrupts(ief);

    for (;;) {
        /*
         * Query number of byte available in MP3 buffer.
         */
        ief = VsPlayerInterrupts(0);
        mp3buf = (u_char *)NutSegBufWriteRequest(&rbytes);
        VsPlayerInterrupts(ief);

        /* 
         * Read data directly into the MP3 buffer. 
         */
        if (rbytes) {
            if ((got = _read(fd, mp3buf, rbytes)) > 0) {
                ief = VsPlayerInterrupts(0);
                mp3buf = (u_char *)NutSegBufWriteCommit(got);
                VsPlayerInterrupts(ief);
            } else {
                break;
            }
        }

        /*
         * If the player is not running, kick it.
         */
        if (VsGetStatus() != VS_STATUS_RUNNING) {
            VsPlayerKick();
        }

        /*
         * Allow background threads to take over.
         */
        NutThreadYield();
    }

    _close(fd);

    /* 
     * Flush decoder and wait until finished. 
     */
    VsPlayerFlush();
    while (VsGetStatus() != VS_STATUS_EMPTY) {
        NutSleep(1);
    }
    printf("\n%u seconds, %u bytes free\n", VsPlayTime(), (u_int)NutHeapAvailable());

    /*
     * Reset the decoder. 
     */
    VsPlayerReset(0);

    return 0;
}

THREAD(VsFeeder, arg)
{
    DIR *dir;
    char *cp;

    /*
     * Initialize the MP3 buffer. The NutSegBuf routines provide a global
     * system buffer, which works with banked and non-banked systems.
     */
    if (NutSegBufInit(32768) == 0) {
        puts("NutSegBufInit: Fatal error");
    }


    /*
     * Play the MP3 files in an endless loop. For each file set the volume
     * of the left and right channel, call the local routine PlayMp3File()
     * and sleep one second before playing the next sound file.
     */
    VsPlayerReset(0);
    for (;;) {
        dir = opendir("PHAT0:/");
        if (dir) {
            struct dirent *dep;

            while((dep = readdir(dir)) != NULL) {
                if ((cp = strrchr(dep->d_name, '.')) != NULL && stricmp(cp, ".mp3") == 0) {
                    char *name = malloc(sizeof(MUSIC_ROOT_DIR) + strlen(dep->d_name));

                    Media2DisplayWriteRow(1, dep->d_name);
                    strcpy(name, MUSIC_ROOT_DIR);
                    strcat(name, dep->d_name);
                    VsSetVolume(32, 32);
                    PlayMp3File(name);

                    free(name);
                }
            }
            closedir(dir);
        }
    }
}

void Media2Backlit(ureg_t on)
{
#if defined(__AVR__)
    sbi(DDRB, 7);
    if (on) {
        sbi(PORTB, 7);
    }
    else {
        cbi(PORTB, 7);
    }
#else
    outr(PIO_PER, _BV(7));
    outr(PIO_OER, _BV(7));
    if (on) {
        outr(PIO_SODR, _BV(7));
    }
    else {
        outr(PIO_CODR, _BV(7));
    }
#endif
}

/*
 * Main application routine. 
 */
int main(void)
{
    u_long baud = 115200;
    int fd;
    int volid;

    Media2Backlit(0);

    /*
     * Register the UART device, open it, assign stdout to it and set 
     * the baudrate.
     */
    NutRegisterDevice(&DEV_DEBUG, 0, 0);
    freopen(DEV_DEBUG_NAME, "w", stdout);
    _ioctl(_fileno(stdout), UART_SETSPEED, &baud);
    puts("\nMedianut VS10XX Decoder Demo 3");

    /*
     * Initialize the MP3 decoder hardware.
     */
    if (VsPlayerInit()) {
        puts("VsPlayer: Fatal error");
        for(;;);
    }
    VsPlayerInterrupts(0);
    VsPlayerThrottle(1);

    /* Initialize the file system. */
    printf("Register file system...");
    if (NutRegisterDevice(&devPhat0, 0, 0)) {
        puts("failed");
        for (;;);
    }
    puts("OK");

    /* Register block device. */
    printf("Register block device...");
    if (NutRegisterDevice(&devNplMmc0, 0, 0)) {
        puts("failed");
        for (;;);
    }
    puts("OK");

    /* Mount partition. */
    printf("Mounting partition...");
    if ((volid = _open("MMC0:1/PHAT0", _O_RDWR | _O_BINARY)) == -1) {
        puts("failed");
        for (;;);
    }
    puts("OK");

    /*
     * Initialize file system.
     */
    NutRegisterDevice(&devUrom, 0, 0);

    /*
     * Open the firmware image file.
     */
    if ((fd = _open(MEDIA2_FIRMWARE_FILE, _O_RDONLY | _O_BINARY)) == -1) {
        printf("Error %d: Failed to open %s.\n", errno, MEDIA2_FIRMWARE_FILE);
    } else {

        printf("Checking firmware...");
        if (Media2Firmware(fd, 0)) {
            printf("not found\nUpdating...");
            _seek(fd, 0, SEEK_SET);
            if (Media2Firmware(fd, 1)) {
                puts("failed");
            } else {
                printf("OK\nVerifying...");
                _seek(fd, 0, SEEK_SET);
                if (Media2Firmware(fd, 0)) {
                    puts("failed");
                } else {
                    puts("OK");
                }
            }
        } else {
            puts("OK");
        }
        _close(fd);
    }

    VsPlayerThrottle(0);

    for (;;) {
        NutSleep(200);
        if (Media2DisplayMode(LCD_MODE_2X16) == 0) {
            break;
        }
        puts("Error: Can't LCD mode");
    }

    NutSleep(200);
    Media2DisplayWriteRow(0, "  Medianut MP3");
    Media2DisplayWriteRow(1, "      Demo");
    Media2Backlit(1);

    NutThreadCreate("vsfeed", VsFeeder, 0, 1024);
    for (;; ) {
        NutSleep(5000);
    }
    return 0;
}
